/*
	bfloop, an interpreter for the BLooP and FLooP languages.

	David Fifield <david@bamsoftware.com>
*/

#include <errno.h>
#include <getopt.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "error.h"
#include "eval.h"
#include "parse.h"

/* How to identify stdin in output messages. */
static const char *STDIN_NAME = "<stdin>";

static int run_file(bfloop_state *state, FILE *fp, const char *filename, bfloop_node **trees);
static const char *filename_extension(const char *filename);
static void print_usage(void);

static const struct option long_options[] = {
	{ "bloop", no_argument, NULL, 0 },
	{ "floop", no_argument, NULL, 0 },
	{ "help", no_argument, NULL, 0 },
	{ "version", no_argument, NULL, 0 },
	{ 0, 0, 0, 0 },
};

/* The global program invocation name (argv[0]), used for log messages. */
const char *program_name = NULL;

/* Check that they didn't ask for --bloop and --floop together. */
#define MODE_SET(mode, new) \
do { \
	if ((mode) != -1 && (mode) != (new)) { \
		printerr("Only one of --bloop and --floop may be given.\n"); \
		exit(1); \
	} \
	(mode) = (new); \
} while (0)

int main(int argc, char *argv[])
{
	bfloop_state state;
	bfloop_node *trees;
	enum lang_mode mode = -1;
	int opt, optindex, ret, i;

	setlocale(LC_ALL, "");

	program_name = argv[0];

	while ((opt = getopt_long(argc, argv, "", long_options, &optindex)) != -1) {
		const char *optname;

		switch (opt) {
		case 0:
			optname = long_options[optindex].name;
			if (strcmp(optname, "bloop") == 0) {
				MODE_SET(mode, BLOOP_MODE);
			} else if (strcmp(optname, "floop") == 0) {
				MODE_SET(mode, FLOOP_MODE);
			} else if (strcmp(optname, "help") == 0) {
				print_usage();
				exit(0);
			} else if (strcmp(optname, "version") == 0) {
				printf("%s %s\n", PACKAGE, PACKAGE_VERSION);
				exit(0);
			} else {
				internal_error("Unhandled long option.");
			}
			break;
		case '?':
		default:
			/* getopt_long prints an error message. */
			exit(1);
			break;
		}
	}

	/* If no mode was asked for, use FLooP mode if any of the files has a
	   name ending in ".floop", otherwise use BLooP mode. */
	if (mode == -1) {
		mode = BLOOP_MODE;
		for (i = optind; i < argc; i++) {
			if (strcmp(filename_extension(argv[i]), ".floop") == 0)
				mode = FLOOP_MODE;
		}
	}
	bfloop_set_lang_mode(mode);

	bfloop_state_init(&state);

	/* Link together each file we parse so we can free them all together. */
	trees = NULL;

	ret = 0;

	if (optind == argc) {
		ret = run_file(&state, stdin, STDIN_NAME, &trees);
		if (ret != 0)
			printerr("Error running %s.\n", STDIN_NAME);
	}

	for (i = optind; i < argc; i++) {
		FILE *fp;
		const char *filename;

		filename = argv[i];
		if (strcmp(filename, "-") == 0) {
			fp = stdin;
			filename = STDIN_NAME;
		} else {
			fp = fopen(filename, "rb");
		}
		if (fp == NULL) {
			printerr("Can't open input file \"%s\": %s.\n",
				filename, strerror(errno));
			exit(1);
		}
		ret = run_file(&state, fp, filename, &trees);
		fclose(fp);
		if (ret != 0) {
			printerr("Error running %s.\n", filename);
			break;
		}
	}

	bfloop_tree_free(trees);
	bfloop_state_free(&state);

	return ret == 0 ? 0 : 1;
}

static int run_file(bfloop_state *state, FILE *fp, const char *filename, bfloop_node **trees)
{
	bfloop_node *tree, *p;
	int ret;

	bfloop_set_filename(filename);

	/* Be interactive on a TTY. */
	if (fp == stdin && isatty(0))
		bfloop_set_parse_mode(STATEMENT_MODE);
	else
		bfloop_set_parse_mode(FILE_MODE);

	ret = 0;
	for (;;) {
		tree = bfloop_parse(fp);
		if (tree == NULL)
			/* yyparse prints an error message. */
			return -1;

		/* Link this into the list of all trees parsed. */
		internal_assert(tree->siblings == NULL);
		if (*trees == NULL) {
			*trees = tree;
		} else {
			for (p = *trees; p->siblings != NULL; p = p->siblings)
				;
			p->siblings = tree;
		}

		if (tree->children == NULL)
			/* Signals end of file. */
			break;

		ret = bfloop_eval(state, tree);
	}

	return ret;
}

static const char *filename_extension(const char *filename)
{
	const char *p;

	p = strrchr(filename, '.');
	if (p == NULL)
		p = strchr(filename, '\0');

	return p;
}

static void print_usage(void)
{
	printf("Usage: %s [OPTION] [FILE]...\n", program_name);
	printf(
"Run a BLooP or FLooP program.\n"
"The program is assembled by concatenating each FILE. Use --bloop or\n"
"--floop to request a specific language mode. The default is BLooP\n"
"(MU-LOOP is prohibited). If any FILE has the extension “.floop”, FLooP\n"
"mode is activated, unless --bloop is used. The encoding of each file\n"
"must be UTF-8.\n"
"\n"
"  --bloop    BLooP mode (default, no MU-LOOP)\n"
"  --floop    FLooP mode\n"
"  --help     display this help and exit\n"
"  --version  display the version number and exit\n"
"\n"
"If no FILE is given, read standard input. - for a FILE means to read\n"
"standard input in that place.\n"
	);
}
