%{
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>

#include "bigint.h"
#include "error.h"
#include "parse.h"

int yylex(void);
void yyerror(const char *s);

/* Where we store the completed parse tree. */
bfloop_node *bfloop_parse_tree = NULL;

/* Whether we're parsing BLooP or FLooP. */
enum lang_mode bfloop_lang_mode = BLOOP_MODE;

/* Whether we've seen a function call. No procedure definitions are allowed
   after that. */
static int seen_call = 0;

/* Track blocks within procedures to give an error on duplicate labels. */
static const bfloop_node **blocks = NULL;
static unsigned int num_blocks = 0;
static unsigned int capacity_blocks = 0;

static int printerr_loc(const YYLTYPE *loc, const char *fmt, ...)
{
	va_list va;
	int ret;

	printerr("%s:%d:%d: ", loc->filename, loc->first_line, loc->first_column);
	va_start(va, fmt);
	ret = vfprintf(stderr, fmt, va);
	va_end(va);

	return ret;
}
%}

%error-verbose

%token DEFINE PROCEDURE QUOTE BLOCK BEGIN END
%token LOOP MU_LOOP AT MOST TIMES IF THEN
%token QUIT ABORT
%nonassoc '+' MUL
%nonassoc '=' '<' '>'
%nonassoc AND
%nonassoc GETS
%token CELL OUTPUT
%token INT IDENT NO YES

%%

toplevel: opt_toplevel_stmt_list
	{
		$$ = bfloop_node_new(NODE_TOPLEVEL);
		if ($$ == NULL)
			YYERROR;
		$$->children = $1;

		bfloop_parse_tree = $$;
	}

opt_toplevel_stmt_list:	{ $$ = NULL; }
	| toplevel_stmt '.' opt_toplevel_stmt_list
	{
		$$ = $1;
		$$->siblings = $3;
		bfloop_tree_free($2);
	}

toplevel_stmt: procedure_def
	| function_call
	{
		$$ = $1;
		seen_call = 1;
	}

procedure_def:
	DEFINE PROCEDURE QUOTE IDENT QUOTE '[' param_list ']' ':' block
	{
		if (seen_call) {
			printerr_loc(&@1, "''%ls'' follows a call.\n", $4->v.s);
			YYERROR;
		}

		$$ = $4;
		$$->type = NODE_PROCEDURE;
		$$->children = $10;
		$$->children->siblings = $7;

		/* We're done with the procedure. Clear the blocks list. */
		free(blocks);
		blocks = NULL;
		num_blocks = 0;
		capacity_blocks = 0;

		bfloop_tree_free($1);
		bfloop_tree_free($2);
		bfloop_tree_free($3);
		bfloop_tree_free($5);
		bfloop_tree_free($6);
		bfloop_tree_free($8);
		bfloop_tree_free($9);
	}

/* A parameter list may not be empty, or else the shortest Blue program on page
   419 would be shorter. */
param_list: IDENT	{ $$ = $1; $$->type = NODE_IDENT; }
	| IDENT ',' param_list
	{
		$$ = $1;
		$$->type = NODE_IDENT;
		$$->siblings = $3;

		bfloop_tree_free($2);
	}

block_or_statement: block | statement

block: BLOCK INT ':' BEGIN opt_statement_list BLOCK INT ':' END
	{
		unsigned int i;

		/* Check that start and end labels match. */
		if (!bigint_equal(&$2->v.i, &$7->v.i)) {
			wchar_t *s, *t;
			s = bigint_render(&$7->v.i);
			internal_assert(s != NULL);
			t = bigint_render(&$2->v.i);
			internal_assert(t != NULL);
			printerr_loc(&@6, "BLOCK %ls does not match BLOCK %ls at %s:%d:%d.\n",
				s, t, @1.filename, @1.first_line, @1.first_column);
			free(s);
			free(t);
			YYERROR;
		}

		/* Check that the block label is unique in this procedure. */
		for (i = 0; i < num_blocks; i++) {
			if (bigint_equal(&$2->v.i, &blocks[i]->v.i)) {
				wchar_t *s;
				s = bigint_render(&$2->v.i);
				internal_assert(s != NULL);
				printerr_loc(&@1, "BLOCK label %ls is used more than once.\n", s);
				free(s);
				YYERROR;
			}
		}

		$$ = $2;
		$$->type = NODE_BLOCK;
		$$->children = $5;

		/* Add to the list of seen block labels. */
		if (num_blocks >= capacity_blocks) {
			capacity_blocks = (num_blocks + 1) * 2;
			blocks = realloc(blocks, sizeof(*blocks) * capacity_blocks);
			internal_assert(blocks != NULL);
		}
		blocks[num_blocks++] = $$;

		bfloop_tree_free($1);
		bfloop_tree_free($3);
		bfloop_tree_free($4);
		bfloop_tree_free($6);
		bfloop_tree_free($7);
		bfloop_tree_free($8);
		bfloop_tree_free($9);
	}

statement: assign_stmt
	| quit_stmt
	| abort_stmt
	| if_stmt
	| loop_stmt
	| mu_loop_stmt

opt_statement_list: { $$ = NULL; }
	| statement ';' opt_statement_list
	{
		$$ = $1;
		$$->siblings = $3;

		bfloop_tree_free($2);
	}

assign_stmt: lvalue GETS expr
	{
		$$ = $2;
		$$->type = NODE_ASSIGN;
		$$->children = $1;
		$$->children->siblings = $3;
	}

/* Page 411 says that only OUTPUT and cell references may be assigned to. In
   other words you're not allowed to modify procedure arguments. */
lvalue: OUTPUT	{ $$ = $1; $$->type = NODE_OUTPUT; }
	| cell_ref;

quit_stmt: QUIT BLOCK INT
	{
		$$ = $3;
		$$->type = NODE_QUIT;

		bfloop_tree_free($1);
		bfloop_tree_free($2);
	}

abort_stmt: ABORT LOOP INT
	{
		$$ = $3;
		$$->type = NODE_ABORT;

		bfloop_tree_free($1);
		bfloop_tree_free($2);
	}


if_stmt: IF cond ',' THEN ':' block_or_statement
	{
		$$ = $1;
		$$->type = NODE_IF;
		$$->children = $2;
		$$->children->siblings = $6;

		bfloop_tree_free($3);
		bfloop_tree_free($4);
		bfloop_tree_free($5);
	}

cond: expr
	| expr '=' expr
	{
		$$ = $2;
		$$->type = NODE_EQUALS;
		$$->children = $1;
		$$->children->siblings = $3;
	}
	| expr '<' expr
	{
		$$ = $2;
		$$->type = NODE_LESS_THAN;
		$$->children = $1;
		$$->children->siblings = $3;
	}
	| expr '>' expr
	{
		$$ = $2;
		$$->type = NODE_GREATER_THAN;
		$$->children = $1;
		$$->children->siblings = $3;
	}
	| '{' cond AND cond '}'
	{
		$$ = $3;
		$$->type = NODE_AND;
		$$->children = $2;
		$$->children->siblings = $4;

		bfloop_tree_free($1);
		bfloop_tree_free($5);
	}

loop_stmt: LOOP at_most_opt loop_count TIMES ':' block
	{
		$$ = $1;
		$$->type = NODE_LOOP;
		$$->children = $3;
		$$->children->siblings = $6;

		bfloop_tree_free($4);
		bfloop_tree_free($5);
	}

mu_loop_stmt: MU_LOOP ':' block
	{
		if (bfloop_lang_mode != FLOOP_MODE) {
			printerr_loc(&@1, "MU-LOOP is not allowed in BLooP mode; use --floop.\n");
			YYERROR;
		}

		$$ = $1;
		$$->type = NODE_MU_LOOP;
		$$->children = $3;

		bfloop_tree_free($2);
	}

at_most_opt:	{ $$ = NULL; }
	| AT MOST
	{
		$$ = NULL;
		bfloop_tree_free($1);
		bfloop_tree_free($2);
	}

loop_count: expr

expr: val
	| val '+' val
	{
		$$ = $2;
		$$->type = NODE_ADD;
		$$->children = $1;
		$$->children->siblings = $3;
	}
	| val MUL val
	{
		$$ = $2;
		$$->type = NODE_MUL;
		$$->children = $1;
		$$->children->siblings = $3;
	}

val: INT		{ $$ = $1; $$->type = NODE_INT; }
	| IDENT		{ $$ = $1; $$->type = NODE_IDENT; }
	| OUTPUT	{ $$ = $1; $$->type = NODE_OUTPUT; }
	| NO		{ $$ = $1; $$->type = NODE_NO; }
	| YES		{ $$ = $1; $$->type = NODE_YES; }
	| cell_ref
	| function_call

cell_ref: CELL '(' INT ')'
	{
		$$ = $3;
		$$->type = NODE_CELL_REF;

		bfloop_tree_free($1);
		bfloop_tree_free($2);
		bfloop_tree_free($4);
	}

function_call: IDENT '[' arg_list ']'
	{
		$$ = $1;
		$$->type = NODE_CALL;
		$$->children = $3;

		bfloop_tree_free($2);
		bfloop_tree_free($4);
	}

arg_list: val
	| val ',' arg_list
	{
		$$ = $1;
		$$->siblings = $3;

		bfloop_tree_free($2);
	}

%%

void yyerror(const char *msg)
{
	printerr_loc(&yylloc, "%s.\n", msg);
}
