/*
Multiple-precision nonnegative integers. The artihmetic algorithms are
described in section 4.3.1 "The Classical Algorithms" of The Art of
Computer Programming. The render and parse functions are applications of
radix conversion, covered in section 4.4.

Each integer is a linked list of unsigned long digits. Less significant
digits are closer to the head of the list. Zero is represented by a
single zero node. The lists returned by these functions always have a
nonzero node in the highest-order position, except in the case of zero.
*/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <wchar.h>

#include "bigint.h"

enum {
	ULONG_BITS = (sizeof(unsigned long) * CHAR_BIT),
};
static const unsigned long HIGH_BIT = (1UL << (ULONG_BITS - 1));

static struct bigint_node *newnode(void)
{
	struct bigint_node *node;

	node = malloc(sizeof(struct bigint_node));
	if (node == NULL)
		return NULL;
	node->v = 0UL;
	node->next = NULL;

	return node;
}

/* Trim any high-order zero nodes. */
static bigint *normalize(bigint *n)
{
	struct bigint_node *p, *nonzero;

	p = nonzero = *n;
	while (p != NULL) {
		if (p->v != 0UL)
			nonzero = p;
		p = p->next;
	}
	bigint_free(&nonzero->next);
	nonzero->next = NULL;

	return n;
}

int bigint_zero(bigint *n)
{
	*n = newnode();
	if (*n == NULL)
		return 0;
	(*n)->v = 0UL;

	return 1;
}

int bigint_parse(bigint *n, const wchar_t *s)
{
	if (!bigint_zero(n))
		return 0;

	for ( ; *s != L'\0'; s++) {
		unsigned long long mul;
		unsigned long carry;
		struct bigint_node *p, **prev;

		/* Find the digit value of this character. */
		carry = *s - L'0';
		if (carry >= 10) {
			bigint_free(n);
			return 0;
		}

		/* Multiply n by 10, add digit. */
		prev = n;
		for (p = *n; p != NULL; p = p->next) {
			mul = (unsigned long long) p->v * 10UL + carry;
			p->v = (unsigned long) mul;
			carry = mul >> ULONG_BITS;
			prev = &p->next;
		}
		if (carry != 0) {
			struct bigint_node *node = newnode();
			if (node == NULL) {
				bigint_free(n);
				return 0;
			}
			node->v = carry;
			*prev = node;
		}
	}
	normalize(n);

	return 1;
}

/* Divide the number starting at p in place and return the remainder. This
   algorithm is a solution to exercise 16 (short division) in section 4.3.1 of
   The Art of Computer Programming. */
static unsigned long divide(struct bigint_node *p, unsigned long divisor)
{
	unsigned long long r;

	if (p == NULL)
		return 0UL;

	/* Get the remainder from the next higher digit. */
	r = divide(p->next, divisor);
	/* Shift, add it to this digit, and divide. */
	r = (r << ULONG_BITS) + p->v;
	p->v = (unsigned long) (r / divisor);

	if (p->next != NULL && p->next->next == NULL && p->next->v == 0UL) {
		bigint_free(&p->next);
		p->next = NULL;
	}

	return (unsigned long) (r % divisor);
}

wchar_t *bigint_render(const bigint *n)
{
	const wchar_t *DIGITS = L"0123456789";
	struct bigint_node *scratch;
	wchar_t *buf = NULL, *tmp;
	size_t len = 0, capacity = 0;
	unsigned int i;

	/* Make a copy of the number. Repeatedly divide it by 10 and add each
	   remainder digit to the buffer. */
	if (!bigint_copy(&scratch, n))
		return NULL;
	do {
		unsigned long remainder;

		remainder = divide(scratch, 10UL);
		if (len + 1 > capacity) {
			capacity = (len + 1) * 2;
			tmp = realloc(buf, (capacity + 1) * sizeof(*buf));
			if (tmp == NULL) {
				free(buf);
				bigint_free(&scratch);
				return NULL;
			}
			buf = tmp;
		}
		buf[len++] = DIGITS[remainder];
	} while (!bigint_is_zero(&scratch));
	bigint_free(&scratch);
	buf[len] = L'\0';

	/* Trim to size. */
	tmp = realloc(buf, (len + 1) * sizeof(*buf));
	if (tmp == NULL) {
		free(buf);
		return NULL;
	}
	buf = tmp;

	/* Digits are produced in reverse order, so reverse them again. */
	for (i = 0; i < len / 2; i++) {
		wint_t t = buf[i];
		buf[i] = buf[len - i - 1];
		buf[len - i - 1] = t;
	}

	return buf;
}

void bigint_free(bigint *n)
{
	struct bigint_node *p, *next;

	for (p = *n; p != NULL; p = next) {
		next = p->next;
		free(p);
	}
}

int bigint_equal(const bigint *a, const bigint *b)
{
	return bigint_compare(a, b) == 0;
}

int bigint_is_zero(const bigint *n)
{
	return (*n)->v == 0UL && (*n)->next == NULL;
}

int bigint_copy(bigint *to, const bigint *from)
{
	const struct bigint_node *p;
	struct bigint_node **prev;

	*to = NULL;
	prev = to;
	for (p = *from; p != NULL; p = p->next) {
		struct bigint_node *node = newnode();
		if (node == NULL) {
			bigint_free(to);
			return 0;
		}
		node->v = p->v;
		*prev = node;
		prev = &node->next;
	}

	return 1;
}

/* Subtract 1 from n. */
int bigint_dec(bigint *n)
{
	struct bigint_node *p;

	if (bigint_is_zero(n))
		return 0;

	p = *n;
	for (;;) {
		p->v--;
		if (p->v != (unsigned long) -1)
			/* Break if no carry. */
			break;
	}
	normalize(n);

	return 1;
}

int bigint_compare(const bigint *a, const bigint *b)
{
	const struct bigint_node *p, *q;
	int d;

	d = 0;
	for (p = *a, q = *b; p != NULL && q != NULL; p = p->next, q = q->next) {
		if (p->v < q->v)
			d = -1;
		else if (p->v > q->v)
			d = 1;
	}
	if (p == NULL && q == NULL)
		return d;
	else if (p == NULL)
		return -1;
	else /* q == NULL */
		return 1;
}

int bigint_add(bigint *sum, const bigint *a, const bigint *b)
{
	const struct bigint_node *p, *q;
	struct bigint_node **prev;
	unsigned long carry = 0UL;

	*sum = NULL;
	prev = sum;
	p = *a;
	q = *b;
	while (p != NULL || q != NULL) {
		struct bigint_node *node;
		unsigned long pv, qv;

		node = newnode();
		if (node == NULL) {
			free(*sum);
			return 0;
		}

		if (p == NULL) {
			pv = 0UL;
		} else {
			pv = p->v;
			p = p->next;
		}
		if (q == NULL) {
			qv = 0UL;
		} else {
			qv = q->v;
			q = q->next;
		}
		node->v = pv + qv + carry;
		carry = (pv & HIGH_BIT) && (qv & HIGH_BIT);

		*prev = node;
		prev = &node->next;
	}

	return 1;
}

int bigint_mul(bigint *product, const bigint *a, const bigint *b)
{
	const struct bigint_node *p, *q;
	struct bigint_node *r;

	if (!bigint_zero(product))
		return 0;

	r = *product;
	for (q = *b; q != NULL; q = q->next) {
		unsigned long carry = 0UL;
		struct bigint_node *tmp;

		tmp = r;
		for (p = *a; p != NULL; p = p->next) {
			unsigned long long mul;

			mul = (unsigned long long) p->v * q->v + r->v + carry;
			r->v = (unsigned long) mul;
			carry = mul >> ULONG_BITS;
			if (r->next == NULL) {
				struct bigint_node *node = newnode();
				if (node == NULL) {
					bigint_free(product);
					return 0;
				}
				r->next = node;
				r = node;
			}
		}
		r->v = carry;
		r = tmp->next;
	}
	normalize(product);

	return 1;
}
