#include <stdio.h>
#include <stdlib.h>
#include <wchar.h>

#include "error.h"
#include "eval.h"
#include "parse.h"
#include "tree.h"

static void value_free(bfloop_value *value)
{
	if (value->type == BFLOOP_INT)
		bigint_free(&value->i);
}

static int value_copy(bfloop_value *to, const bfloop_value *from)
{
	to->type = from->type;
	if (from->type == BFLOOP_INT)
		return bigint_copy(&to->i, &from->i);

	return 1;
}

static wchar_t *value_to_string(bfloop_value *value)
{
	wchar_t *template, *s;

	if (value->type == BFLOOP_INT)
		return bigint_render(&value->i);

	template = NULL;
	if (value->type == BFLOOP_NONE)
		template = L"NONE";
	else if (value->type == BFLOOP_NO)
		template = L"NO";
	else if (value->type == BFLOOP_YES)
		template = L"YES";
	internal_assert(template != NULL);

	s = malloc(sizeof(wchar_t) * (wcslen(template) + 1));
	if (s == NULL)
		return NULL;
	wcscpy(s, template);

	return s;
}

static int state_push(bfloop_state *state, bfloop_stack_frame *frame);
static void state_pop(bfloop_state *state);

void bfloop_state_init(bfloop_state *state)
{
	state->num_procs = 0;
	state->procs = NULL;
	state->stack = NULL;
	state->quit_abort_node = NULL;
}

void bfloop_state_free(bfloop_state *state)
{
	int i;

	for (i = 0; i < state->num_procs; i++)
		free(state->procs[i].params);
	free(state->procs);
	while(state->stack != NULL)
		state_pop(state);
}

static int proc_name_is_test(const wchar_t *name)
{
	size_t len;

	len = wcslen(name);
	internal_assert(len > 0);

	return name[len - 1] == L'?';
}

static bfloop_stack_frame *bfloop_state_newframe(bfloop_state *state, unsigned int proc_index)
{
	bfloop_stack_frame *frame;

	frame = malloc(sizeof(*frame));
	if (frame == NULL)
		return NULL;

	frame->is_test = proc_name_is_test(state->procs[proc_index].name);

	/* The default value of OUTPUT depends on whether this is a test (name
	   ends in '?') or a procedure. */
	if (frame->is_test) {
		frame->output.type = BFLOOP_NO;
	} else {
		frame->output.type = BFLOOP_INT;
		if (!bigint_zero(&frame->output.i)) {
			free(frame);
			return 0;
		}
	}

	frame->syms = NULL;

	frame->proc_index = proc_index;

	frame->next = NULL;

	return frame;
}

static void bfloop_stack_frame_free(bfloop_stack_frame *frame)
{
	value_free(&frame->output);
	symtab_free(frame->syms);
	free(frame);
}

static int state_push(bfloop_state *state, bfloop_stack_frame *frame)
{
	frame->next = state->stack;
	state->stack = frame;

	return 1;
}

static void state_pop(bfloop_state *state)
{
	bfloop_stack_frame *frame;

	internal_assert(state->stack != NULL);
	frame = state->stack;
	state->stack = frame->next;

	bfloop_stack_frame_free(frame);
}

static void frame_setsym(bfloop_stack_frame *frame, const wchar_t *name, const bigint *value)
{
	internal_assert(frame != NULL);
	frame->syms = symtab_set(frame->syms, name, value);
	internal_assert(frame->syms != NULL);
}

static void state_setsym(bfloop_state *state, const wchar_t *name, const bigint *value)
{
	frame_setsym(state->stack, name, value);
}

static int state_getsym(bfloop_state *state, const wchar_t *name, bigint *value)
{
	const bigint *p;

	if (state->stack == NULL)
		/* Means we're at the top level, with no variable bindings. */
		return 0;
	p = symtab_lookup(state->stack->syms, name);
	if (p == NULL)
		return 0;

	internal_assert(bigint_copy(value, p));
	return 1;
}

/* Change a bigint into a string go it can be used to index a symtab. */
static wchar_t *make_cell_ref_string(const bigint *addr)
{
	wchar_t *buf, *s;
	size_t len;

	s = bigint_render(addr);
	internal_assert(s != NULL);
	len = wcslen(L"CELL(") + wcslen(s) + wcslen(L")") + 1;
	buf = malloc(sizeof(wchar_t) * len);
	internal_assert(buf != NULL);
	*buf = L'\0';
	wcscat(buf, L"CELL(");
	wcscat(buf, s);
	wcscat(buf, L")");
	free(s);

	return buf;
}

static void state_setcell(bfloop_state *state, const bigint *addr, const bigint *value)
{
	wchar_t *name = make_cell_ref_string(addr);
	state_setsym(state, name, value);
	free(name);
}

static void state_getcell(bfloop_state *state, const bigint *addr, bigint *value)
{
	wchar_t *name = make_cell_ref_string(addr);
	if (!state_getsym(state, name, value))
		internal_assert(bigint_zero(value));
	free(name);
}

static int state_lookup_proc(bfloop_state *state, const wchar_t *name)
{
	int i;

	for (i = 0; i < state->num_procs; i++) {
		if (wcscmp(name, state->procs[i].name) == 0)
			return i;
	}

	return -1;
}

static int state_register_proc(bfloop_state *state, const bfloop_node *tree)
{
	bfloop_proc proc;
	bfloop_node *param;
	int i, j;

	proc.name = tree->v.s;

	/* Check the procedure name against previous procedures and parameters. */
	for (i = 0; i < state->num_procs; i++) {
		if (wcscmp(proc.name, state->procs[i].name) == 0) {
			printerr("Duplicate procedure name ''%ls''.\n", proc.name);
			return 0;
		}
		for (j = 0; j < state->procs[i].num_params; j++) {
			if (wcscmp(proc.name, state->procs[i].params[j]) == 0) {
				printerr("Procedure ''%ls'' has the same name as a parameter of procedure ''%ls''.\n", proc.name, state->procs[i].name);
				return 0;
			}
		}
	}

	/* Check the parameters against previous procedure names and parameters. */
	internal_assert(tree->children != NULL);
	proc.num_params = 0;
	for (param = tree->children->siblings; param != NULL; param = param->siblings) {
		const bfloop_node *other;
		internal_assert(param->type == NODE_IDENT);
		if (state_lookup_proc(state, param->v.s) != -1
		    || wcscmp(proc.name, param->v.s) == 0) {
			printerr("Parameter %ls of procedure ''%ls'' has the same name as a procedure.\n", param->v.s, proc.name);
			return 0;
		}
		/* Check for other parameters with the same name. */
		for (other = param->siblings; other != NULL; other = other->siblings) {
			if (other->type == NODE_IDENT && wcscmp(other->v.s, param->v.s) == 0) {
				printerr("Parameter %ls of procedure ''%ls'' is repeated.\n", param->v.s, proc.name);
				return 0;
			}
		}
		proc.num_params++;
	}
	internal_assert(proc.num_params > 0);
	proc.params = malloc(sizeof(*proc.params) * proc.num_params);
	internal_assert(proc.params != NULL);
	i = 0;
	for (param = tree->children->siblings; param != NULL; param = param->siblings)
		proc.params[i++] = param->v.s;

	proc.block = tree->children;

	state->procs = realloc(state->procs, sizeof(*state->procs) * (state->num_procs + 1));
	internal_assert(state->procs != NULL);
	state->procs[state->num_procs++] = proc;

	return 1;
}

static int node_num_children(const bfloop_node *node) {
	int n = 0;

	for (node = node->children; node != NULL; node = node->siblings)
		n++;

	return n;
}

/* Magic return codes indicating special behavior. They must be less than 0. */
enum {
	BFLOOP_ERROR = -1,
	BFLOOP_QUIT = -2,
	BFLOOP_ABORT = -3,
};

static int eval(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_toplevel(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_procedure(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_call(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_block(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_if(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_quit(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_abort(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_loop(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_mu_loop(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_assign(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_mul(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_add(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_and(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_equals(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_less_than(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_greater_than(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_output(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_cell_ref(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_ident(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_int(bfloop_state *, const bfloop_node *, bfloop_value *);
static int eval_yes(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue);
static int eval_no(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue);

int bfloop_eval(bfloop_state *state, const bfloop_node *tree)
{
	return eval(state, tree, NULL);
}

static int eval(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	if (tree == NULL)
		return 1;

	switch (tree->type) {
	case NODE_TOPLEVEL:
		return eval_toplevel(state, tree, rvalue);
	case NODE_PROCEDURE:
		return eval_procedure(state, tree, rvalue);
	case NODE_CALL:
		return eval_call(state, tree, rvalue);
	case NODE_BLOCK:
		return eval_block(state, tree, rvalue);
	case NODE_IF:
		return eval_if(state, tree, rvalue);
	case NODE_QUIT:
		return eval_quit(state, tree, rvalue);
	case NODE_ABORT:
		return eval_abort(state, tree, rvalue);
	case NODE_LOOP:
		return eval_loop(state, tree, rvalue);
	case NODE_MU_LOOP:
		return eval_mu_loop(state, tree, rvalue);
	case NODE_ASSIGN:
		return eval_assign(state, tree, rvalue);
	case NODE_MUL:
		return eval_mul(state, tree, rvalue);
	case NODE_ADD:
		return eval_add(state, tree, rvalue);
	case NODE_EQUALS:
		return eval_equals(state, tree, rvalue);
	case NODE_LESS_THAN:
		return eval_less_than(state, tree, rvalue);
	case NODE_GREATER_THAN:
		return eval_greater_than(state, tree, rvalue);
	case NODE_AND:
		return eval_and(state, tree, rvalue);
	case NODE_OUTPUT:
		return eval_output(state, tree, rvalue);
	case NODE_CELL_REF:
		return eval_cell_ref(state, tree, rvalue);
	case NODE_IDENT:
		return eval_ident(state, tree, rvalue);
	case NODE_INT:
		return eval_int(state, tree, rvalue);
	case NODE_YES:
		return eval_yes(state, tree, rvalue);
	case NODE_NO:
		return eval_no(state, tree, rvalue);
	case NODE_NONE:
		internal_error("invalid type for eval: NONE");
		break;
	}

	internal_error("unknown node type");
}

static int eval_toplevel(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	const bfloop_node *p;
	int ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_TOPLEVEL);

	internal_assert(tree->siblings == NULL);

	for (p = tree->children; p != NULL; p = p->siblings) {
		bfloop_value result;

		if ((ret = eval(state, p, &result)) < 0)
			return ret;
		internal_assert(state->stack == NULL);

		/* Print out the return value of toplevel calls. */
		if (p->type == NODE_CALL) {
			wchar_t *s = value_to_string(&result);
			internal_assert(s != NULL);
			printf("%ls\n", s);
			free(s);
		}

		value_free(&result);
	}

	if (rvalue != NULL)
		rvalue->type = BFLOOP_NONE;

	return 0;
}

static int eval_procedure(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_PROCEDURE);

	if (!state_register_proc(state, tree))
		return BFLOOP_ERROR;

	if (rvalue != NULL)
		rvalue->type = BFLOOP_NONE;

	return 0;
}

static int eval_call(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	int proc_index;
	const bfloop_proc *proc;
	const bfloop_node *param;
	bfloop_stack_frame *frame;
	int i, ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_CALL);

	proc_index = state_lookup_proc(state, tree->v.s);
	if (proc_index == -1) {
		printerr("No procedure named ''%ls''.\n", tree->v.s);
		return BFLOOP_ERROR;
	} else if (state->stack != NULL && proc_index == state->stack->proc_index) {
		internal_assert(state->stack != NULL);
		printerr("Procedure ''%ls'' may not call itself.\n",
			state->procs[state->stack->proc_index].name);
		return BFLOOP_ERROR;
	} else if (state->stack != NULL && proc_index > state->stack->proc_index) {
		printerr("Procedure ''%ls'' may not call ''%ls'', defined later.\n",
			state->procs[state->stack->proc_index].name,
			state->procs[proc_index].name);
		return BFLOOP_ERROR;
	}

	proc = &state->procs[proc_index];

	frame = bfloop_state_newframe(state, proc_index);
	internal_assert(frame != NULL);

	for (i = 0, param = tree->children;
	     i < proc->num_params && param != NULL;
	     i++, param = param->siblings) {
		bfloop_value v;

		if ((ret = eval(state, param, &v)) < 0) {
			bfloop_stack_frame_free(frame);
			return ret;
		}
		if (v.type != BFLOOP_INT) {
			wchar_t *s = value_to_string(&v);
			internal_assert(s != NULL);
			printerr("Parameter %ls to call of ''%ls'' has non-number value %ls.\n", proc->params[i], proc->name, s);
			free(s);
			bfloop_stack_frame_free(frame);
			value_free(&v);
			return BFLOOP_ERROR;
		}
		frame_setsym(frame, proc->params[i], &v.i);
		value_free(&v);
	}
	if (i < proc->num_params) {
		printerr("Call to ''%ls'' has too few arguments (expected %d).\n", proc->name, proc->num_params);
		return BFLOOP_ERROR;
	}
	if (param != NULL) {
		printerr("Call to ''%ls'' has too many arguments (expected %d).\n", proc->name, proc->num_params);
		return BFLOOP_ERROR;
	}

	internal_assert(state_push(state, frame));

	ret = eval(state, proc->block, NULL);

	if (ret == BFLOOP_QUIT) {
		wchar_t *s;
		internal_assert(state->quit_abort_node != NULL);
		s = bigint_render(&state->quit_abort_node->v.i);
		internal_assert(s != NULL);
		printerr("QUIT %ls propagated outside call.\n", s);
		free(s);
		return BFLOOP_ERROR;
	} else if (ret == BFLOOP_ABORT) {
		wchar_t *s;
		internal_assert(state->quit_abort_node != NULL);
		s = bigint_render(&state->quit_abort_node->v.i);
		internal_assert(s != NULL);
		printerr("ABORT %ls propagated outside call.\n", s);
		free(s);
		return BFLOOP_ERROR;
	} else if (ret < 0) {
		return ret;
	}

	internal_assert(value_copy(rvalue, &state->stack->output));

	state_pop(state);

	return 0;
}

static int eval_block(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	const bfloop_node *p;
	int ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_BLOCK);

	for (p = tree->children; p != NULL; p = p->siblings) {
		ret = eval(state, p, NULL);
		if (ret == BFLOOP_QUIT
		    && bigint_compare(&tree->v.i, &state->quit_abort_node->v.i) == 0) {
			/* Quit this block. */
			break;
		} else if (ret < 0) {
			return ret;
		}
	}

	if (rvalue != NULL)
		rvalue->type = BFLOOP_NONE;

	return 0;
}

static int eval_if(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	bfloop_value val;
	int v, ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_IF);
	internal_assert(node_num_children(tree) == 2);

	if ((ret = eval(state, tree->children, &val)) < 0)
		return ret;

	if (!(val.type == BFLOOP_NO || val.type == BFLOOP_YES)) {
		wchar_t *s = value_to_string(&val);
		internal_assert(s != NULL);
		printerr("Condition is not NO or YES: %ls.\n", s);
		free(s);
		value_free(&val);
		return BFLOOP_ERROR;
	}

	v = (val.type == BFLOOP_YES);
	value_free(&val);

	if (v) {
		if ((ret = eval(state, tree->children->siblings, NULL)) < 0)
			return ret;
	}

	return 0;
}

static int eval_quit(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_QUIT);
	internal_assert(tree->children == NULL);

	state->quit_abort_node = tree;

	return BFLOOP_QUIT;
}

static int eval_abort(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_ABORT);
	internal_assert(tree->children == NULL);

	state->quit_abort_node = tree;

	return BFLOOP_ABORT;
}

static int eval_loop(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	const bfloop_node *body;
	bfloop_value limit;
	int ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_LOOP);
	internal_assert(node_num_children(tree) == 2);

	if ((ret = eval(state, tree->children, &limit)) < 0)
		return ret;

	if (limit.type != BFLOOP_INT) {
		wchar_t *s = value_to_string(&limit);
		internal_assert(s != NULL);
		printerr("Loop limit is not numeric: %ls.\n", s);
		free(s);
		value_free(&limit);
		return BFLOOP_ERROR;
	}

	body = tree->children->siblings;

	while (!bigint_is_zero(&limit.i)) {
		internal_assert(bigint_dec(&limit.i));
		ret = eval(state, body, NULL);
		if (ret == BFLOOP_ABORT && body->type == NODE_BLOCK
		    && bigint_compare(&body->v.i, &state->quit_abort_node->v.i) == 0) {
			/* Abort this loop. */
			break;
		} else if (ret < 0) {
			value_free(&limit);
			return ret;
		}
	}

	value_free(&limit);

	if (rvalue != NULL)
		rvalue->type = BFLOOP_NONE;

	return 0;
}

static int eval_mu_loop(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	const bfloop_node *body;
	int ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_MU_LOOP);
	internal_assert(node_num_children(tree) == 1);

	body = tree->children;

	for (;;) {
		ret = eval(state, body, NULL);
		if (ret == BFLOOP_ABORT && body->type == NODE_BLOCK
		    && bigint_compare(&body->v.i, &state->quit_abort_node->v.i) == 0) {
			/* Abort this loop. */
			break;
		} else if (ret < 0) {
			return ret;
		}
	}

	if (rvalue != NULL)
		rvalue->type = BFLOOP_NONE;

	return 0;
}

static int eval_assign(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	const bfloop_node *left, *right;
	bfloop_value v;
	int ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_ASSIGN);
	internal_assert(node_num_children(tree) == 2);

	left = tree->children;
	right = tree->children->siblings;

	internal_assert(left->children == NULL);

	if ((ret = eval(state, right, &v)) < 0)
		return ret;

	if (left->type == NODE_CELL_REF) {
		if (v.type != BFLOOP_INT) {
			wchar_t *s = value_to_string(&v);
			internal_assert(s != NULL);
			printerr("Right side is not numeric in cell assignment: %ls.\n", s);
			free(s);
			value_free(&v);
			return BFLOOP_ERROR;
		}
		state_setcell(state, &left->v.i, &v.i);
	} else if (left->type == NODE_OUTPUT) {
		if ((state->stack->output.type == BFLOOP_NO
		     || state->stack->output.type == BFLOOP_YES)
		    && !(v.type == BFLOOP_NO
			 || v.type == BFLOOP_YES)) {
			wchar_t *s = value_to_string(&v);
			internal_assert(s != NULL);
			printerr("Assignment to OUTPUT in test expects NO or YES, not %ls.\n", s);
			free(s);
			value_free(&v);
			return BFLOOP_ERROR;
		}
		if (state->stack->output.type == BFLOOP_INT
		    && v.type != BFLOOP_INT) {
			wchar_t *s = value_to_string(&v);
			internal_assert(s != NULL);
			printerr("Assignment to OUTPUT in procedure expects number, not %ls.\n", s);
			free(s);
			value_free(&v);
			return BFLOOP_ERROR;
		}
		/* OUTPUT already contains a value. */
		value_free(&state->stack->output);
		value_copy(&state->stack->output, &v);
	} else {
		internal_error("invalid node type for lvalue");
	}

	value_free(&v);

	if (rvalue != NULL)
		rvalue->type = BFLOOP_NONE;

	return 0;
}

static int eval_mul(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	bfloop_value a, b;
	int ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_MUL);
	internal_assert(node_num_children(tree) == 2);

	if ((ret = eval(state, tree->children, &a)) < 0)
		return ret;
	if (a.type != BFLOOP_INT) {
		wchar_t *s = value_to_string(&a);
		internal_assert(s != NULL);
		printerr("Value in multiplication is not numeric: %ls.\n", s);
		free(s);
		value_free(&a);
		return BFLOOP_ERROR;
	}

	if ((ret = eval(state, tree->children->siblings, &b)) < 0) {
		value_free(&a);
		return ret;
	}
	if (b.type != BFLOOP_INT) {
		wchar_t *s = value_to_string(&b);
		internal_assert(s != NULL);
		printerr("Value in multiplication is not numeric: %ls.\n", s);
		free(s);
		value_free(&a);
		value_free(&b);
		return BFLOOP_ERROR;
	}

	if (rvalue != NULL) {
		rvalue->type = BFLOOP_INT;
		internal_assert(bigint_mul(&rvalue->i, &a.i, &b.i));
	}

	value_free(&a);
	value_free(&b);

	return 0;
}

static int eval_add(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	bfloop_value a, b;
	int ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_ADD);
	internal_assert(node_num_children(tree) == 2);

	if ((ret = eval(state, tree->children, &a)) < 0)
		return ret;
	if (a.type != BFLOOP_INT) {
		wchar_t *s = value_to_string(&a);
		internal_assert(s != NULL);
		printerr("Value in addition is not numeric: %ls.\n", s);
		free(s);
		value_free(&a);
		return BFLOOP_ERROR;
	}

	if ((ret = eval(state, tree->children->siblings, &b)) < 0) {
		value_free(&a);
		return ret;
	}
	if (b.type != BFLOOP_INT) {
		wchar_t *s = value_to_string(&b);
		internal_assert(s != NULL);
		printerr("Value in addition is not numeric: %ls.\n", s);
		free(s);
		value_free(&a);
		value_free(&b);
		return BFLOOP_ERROR;
	}

	if (rvalue != NULL) {
		rvalue->type = BFLOOP_INT;
		internal_assert(bigint_add(&rvalue->i, &a.i, &b.i));
	}

	value_free(&a);
	value_free(&b);

	return 0;
}

static int eval_and(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	bfloop_value a, b;
	int ret;

	internal_assert(tree != NULL);
	internal_assert(tree->children != NULL);
	internal_assert(node_num_children(tree) == 2);

	if ((ret = eval(state, tree->children, &a)) < 0)
		return ret;
	if (!(a.type == BFLOOP_NO || a.type == BFLOOP_YES)) {
		wchar_t *s = value_to_string(&a);
		internal_assert(s != NULL);
		printerr("Value in AND is not NO or YES: %ls.\n", s);
		free(s);
		value_free(&a);
		return BFLOOP_ERROR;
	}

	if ((ret = eval(state, tree->children->siblings, &b)) < 0) {
		value_free(&a);
		return ret;
	}
	if (!(b.type == BFLOOP_NO || b.type == BFLOOP_YES)) {
		wchar_t *s = value_to_string(&b);
		internal_assert(s != NULL);
		printerr("Value in AND is not NO or YES: %ls.\n", s);
		free(s);
		value_free(&a);
		value_free(&b);
		return BFLOOP_ERROR;
	}

	if (rvalue != NULL) {
		if (a.type == BFLOOP_YES && b.type == BFLOOP_YES)
			rvalue->type = BFLOOP_YES;
		else
			rvalue->type = BFLOOP_NO;
	}

	value_free(&a);
	value_free(&b);

	return 0;
}

/* Evaluate the two children of tree and print an error if they are not both
   numeric. Compare them as with strcmp and store the result in cmp. Returns 0
   on failure. This is used to avoid code duplication in the < and >
   operators. */
static int tree_num_compare(bfloop_state *state, const bfloop_node *tree, int *cmp)
{
	bfloop_value a, b;
	int ret;

	internal_assert(tree != NULL);
	internal_assert(tree->children != NULL);
	internal_assert(node_num_children(tree) == 2);

	if ((ret = eval(state, tree->children, &a)) < 0)
		return ret;
	if (a.type != BFLOOP_INT) {
		wchar_t *s = value_to_string(&a);
		internal_assert(s != NULL);
		printerr("Value in comparison is not numeric: %ls.\n", s);
		free(s);
		value_free(&a);
		return BFLOOP_ERROR;
	}

	if ((ret = eval(state, tree->children->siblings, &b)) < 0) {
		value_free(&a);
		return ret;
	}
	if (b.type != BFLOOP_INT) {
		wchar_t *s = value_to_string(&b);
		internal_assert(s != NULL);
		printerr("Value in comparison is not numeric: %ls.\n", s);
		free(s);
		value_free(&a);
		value_free(&b);
		return BFLOOP_ERROR;
	}

	*cmp = bigint_compare(&a.i, &b.i);

	value_free(&a);
	value_free(&b);

	return 0;
}

static int eval_equals(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	int cmp, ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_EQUALS);

	if ((ret = tree_num_compare(state, tree, &cmp)) < 0)
		return ret;

	if (cmp == 0)
		rvalue->type = BFLOOP_YES;
	else
		rvalue->type = BFLOOP_NO;

	return 0;
}

static int eval_less_than(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	int cmp, ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_LESS_THAN);

	if ((ret = tree_num_compare(state, tree, &cmp)) < 0)
		return ret;

	if (cmp < 0)
		rvalue->type = BFLOOP_YES;
	else
		rvalue->type = BFLOOP_NO;

	return 0;
}

static int eval_greater_than(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	int cmp, ret;

	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_GREATER_THAN);

	if ((ret = tree_num_compare(state, tree, &cmp)) < 0)
		return ret;

	if (cmp > 0)
		rvalue->type = BFLOOP_YES;
	else
		rvalue->type = BFLOOP_NO;

	return 0;
}

static int eval_output(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_OUTPUT);
	internal_assert(tree->children == NULL);

	if (rvalue != NULL)
		internal_assert(value_copy(rvalue, &state->stack->output));

	return 0;
}

static int eval_cell_ref(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_CELL_REF);
	internal_assert(tree->children == NULL);

	if (rvalue != NULL) {
		rvalue->type = BFLOOP_INT;
		state_getcell(state, &tree->v.i, &rvalue->i);
	}

	return 0;
}

static int eval_ident(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_IDENT);
	internal_assert(tree->children == NULL);

	if (rvalue != NULL) {
		rvalue->type = BFLOOP_INT;
		if (!state_getsym(state, tree->v.s, &rvalue->i)) {
			printerr("No parameter named %ls.\n", tree->v.s);
			return BFLOOP_ERROR;
		}
	}

	return 0;
}

static int eval_int(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_INT);
	internal_assert(tree->children == NULL);

	if (rvalue != NULL) {
		rvalue->type = BFLOOP_INT;
		internal_assert(bigint_copy(&rvalue->i, &tree->v.i));
	}

	return 0;
}

static int eval_yes(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_YES);
	internal_assert(tree->children == NULL);

	if (rvalue != NULL)
		rvalue->type = BFLOOP_YES;

	return 0;
}

static int eval_no(bfloop_state *state, const bfloop_node *tree, bfloop_value *rvalue)
{
	internal_assert(tree != NULL);
	internal_assert(tree->type == NODE_NO);
	internal_assert(tree->children == NULL);

	if (rvalue != NULL)
		rvalue->type = BFLOOP_NO;

	return 0;
}
