#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <wctype.h>

#include "error.h"
#include "parse.h"
#include "tree.h"
#include "utf8.h"

int yyparse(void);

extern int yydebug;

/* The global bfloop_node tree assigned in the last step of yyparse. */
extern bfloop_node *bfloop_parse_tree;

/* The name of the file we're currently parsing. */
static const char *bfloop_parse_filename;

/* Whether to return parse results line-by-line or a whole file at a time. */
static enum parse_mode bfloop_parse_mode = FILE_MODE;

/* The stream we're parsing from and one character of pushback. */
static FILE *parse_fp;
static wint_t ungotc = WEOF;

/* Growing character buffer. */
struct buffer {
	wchar_t *s;
	size_t n;	/* Size of allocation. */
	size_t len;	/* Length of string. */
};

static void buffer_init(struct buffer *buffer)
{
	buffer->s = NULL;
	buffer->n = 0;
	buffer->len = 0;
}

/* Add c to buf, growing if necessary. Frees the buffer and returns NULL on
   allocation error. */
static wchar_t *buffer_append(struct buffer *buf, wint_t c)
{
	wchar_t *p;

	while (buf->len + 1 > buf->n) {
		if (buf->n == 0)
			buf->n = 1;
		else
			buf->n *= 2;
		p = realloc(buf->s, sizeof(wchar_t) * buf->n);
		if (p == NULL) {
			free(buf->s);
			return NULL;
		}
		buf->s = p;
	}
	buf->s[buf->len++] = c;

	return buf->s;
}

static wint_t stream_getc(void)
{
	wint_t c;

	if (ungotc != WEOF) {
		c = ungotc;
		ungotc = WEOF;
	} else {
		errno = 0;
		c = utf8_getc(parse_fp);
		if (c == WEOF && errno != 0) {
			printerr("%s:%d:%d: %s.\n", yylloc.filename, yylloc.last_line, yylloc.last_column, strerror(errno));
			exit(2);
		}
	}

	return c;
}

static void stream_ungetc(wint_t c)
{
	internal_assert(ungotc == WEOF);
	ungotc = c;
}

/* Classify a string token either as its specific keyword if it is a keyword or
   else as a generic identifier. */
static int classify_token(wchar_t *s)
{
	const struct {
		wchar_t *s;
		int type;
	} TOKEN_MAP[] = {
		{ L"DEFINE", DEFINE },
		{ L"PROCEDURE", PROCEDURE },
		{ L"BLOCK", BLOCK },
		{ L"BEGIN", BEGIN },
		{ L"END", END },
		{ L"LOOP", LOOP },
		{ L"MU-LOOP", MU_LOOP },
		{ L"AT", AT },
		{ L"MOST", MOST },
		{ L"TIMES", TIMES },
		{ L"IF", IF },
		{ L"THEN", THEN },
		{ L"QUIT", QUIT },
		{ L"ABORT", ABORT },
		{ L"AND", AND },
		{ L"CELL", CELL },
		{ L"OUTPUT", OUTPUT },
		{ L"YES", YES },
		{ L"NO", NO },
	};
	int i;

	for (i = 0; i < sizeof(TOKEN_MAP) / sizeof(TOKEN_MAP[0]); i++) {
		if (wcscmp(s, TOKEN_MAP[i].s) == 0) {
			free(s);
			return TOKEN_MAP[i].type;
		}
	}

	yylval->type = NODE_IDENT;
	yylval->v.s = s;

	return IDENT;
}

/* Only a limited repertoire of 56 characters is allowed. See page 419. */

static int is_whitespace_char(wint_t c)
{
	/* We recognize '\n' and '\t' as synonyms for ' ', however. */
	return c == L' ' || c == L'\n' || c == L'\t';
}

static int is_ident_char(wint_t c)
{
	const wchar_t IDENT_CHARS[] = L"ABCDEFGHIJKLMNOPQRSTUVWXYZ-?";

	return wcschr(IDENT_CHARS, c) != NULL;
}

static int is_ident_initial_char(wint_t c)
{
	return is_ident_char(c) && c != L'-' && c != L'?';
}

static int is_digit_char(wint_t c)
{
	const wchar_t DIGITS[] = L"0123456789";

	return wcschr(DIGITS, c) != NULL;
}

/* Lexical scanner function called by yyparse. */
int yylex(void)
{
	/* Remember whether the last call completed a statement, for
	   STATEMENT_MODE. */
	static int seen_dot = 0;
	struct buffer buf;
	wint_t c;

	/*
	yydebug = 1;
	*/

	yylloc.filename = bfloop_parse_filename;

	if (seen_dot) {
		seen_dot = 0;
		if (bfloop_parse_mode == STATEMENT_MODE)
			/* Return a tree since the previous statement is
			   finished. */
			return 0;
	}

	/* Skip white space. */
	while ((c = stream_getc()) != WEOF && is_whitespace_char(c)) {
		if (c == L'\n') {
			yylloc.last_line++;
			yylloc.last_column = 0;
		} else {
			yylloc.last_column++;
		}
		yylloc.last_line = yylloc.first_line;
		yylloc.last_column = yylloc.first_column;
	}

	yylloc.first_line = yylloc.last_line;
	yylloc.first_column = yylloc.last_column;

	if (c == WEOF)
		return 0;

	yylval = bfloop_node_new(NODE_NONE);
	internal_assert(yylval != NULL);
	yylval->v.s = NULL;

	buffer_init(&buf);

	if (is_ident_initial_char(c)) {
		do {
			internal_assert(buffer_append(&buf, c) != NULL);
			yylloc.last_column++;
		} while ((c = stream_getc()) != WEOF && is_ident_char(c));
		stream_ungetc(c);
		internal_assert(buffer_append(&buf, L'\0') != NULL);
		return classify_token(buf.s);
	}

	if (is_digit_char(c)) {
		do {
			internal_assert(buffer_append(&buf, c) != NULL);
			yylloc.last_column++;
		} while ((c = stream_getc()) != WEOF && is_digit_char(c));
		stream_ungetc(c);
		internal_assert(buffer_append(&buf, L'\0') != NULL);
		internal_assert(bigint_parse(&yylval->v.i, buf.s));
		free(buf.s);
		yylval->type = NODE_INT;
		return INT;
	}

	if (c == L'\'') {
		c = stream_getc();
		if (c != L'\'') {
			stream_ungetc(c);
			return L'\'';
		} else {
			yylloc.last_column++;
			return QUOTE;
		}
	}

	/* We must define symbols for characters whose value is above 0xFF
	   because Bison can't handle them. */
	if (c == L'⇐')
		return GETS;
	else if (c == L'×')
		return MUL;

	if (c == L'.')
		seen_dot = 1;

	return c;
}

void bfloop_set_filename(const char *filename)
{
	bfloop_parse_filename = filename;
}

enum lang_mode bfloop_set_lang_mode(enum lang_mode mode)
{
	/* Defined in bfloop.tab.y. */
	extern enum lang_mode bfloop_lang_mode;
	enum lang_mode old_mode;

	old_mode = bfloop_lang_mode;
	bfloop_lang_mode = mode;

	return old_mode;
}

enum parse_mode bfloop_set_parse_mode(enum parse_mode mode)
{
	enum parse_mode old_mode;

	old_mode = bfloop_parse_mode;
	bfloop_parse_mode = mode;

	return old_mode;
}

bfloop_node *bfloop_parse(FILE *fp)
{
	int ret;

	/* Initialize the tree that will hold the program if parsing is
	   successful. */
	bfloop_parse_tree = NULL;

	parse_fp = fp;

	yylloc.first_line = yylloc.last_line = 1;
	yylloc.first_column = yylloc.last_column = 0;

	ret = yyparse();

	if (ret == 0)
		return bfloop_parse_tree;
	else
		return NULL;
}
