#include <stdio.h>
#include <stdlib.h>

#include "error.h"
#include "tree.h"

bfloop_node *bfloop_node_new(enum bfloop_node_type type)
{
	bfloop_node *node;

	node = malloc(sizeof(*node));
	if (node == NULL)
		return NULL;
	node->type = type;
	node->siblings = NULL;
	node->children = NULL;

	return node;
}

static void bfloop_node_free(bfloop_node *node)
{
	switch (node->type) {
	case NODE_NONE:
	case NODE_NO:
	case NODE_YES:
	case NODE_OUTPUT:
	case NODE_ADD:
	case NODE_MUL:
	case NODE_AND:
	case NODE_EQUALS:
	case NODE_LESS_THAN:
	case NODE_GREATER_THAN:
	case NODE_ASSIGN:
	case NODE_LOOP:
	case NODE_MU_LOOP:
	case NODE_IF:
	case NODE_TOPLEVEL:
		/* Nothing necessary. */
		break;
	case NODE_IDENT:
	case NODE_CALL:
	case NODE_PROCEDURE:
		free(node->v.s);
		break;
	case NODE_INT:
	case NODE_CELL_REF:
	case NODE_QUIT:
	case NODE_ABORT:
	case NODE_BLOCK:
		bigint_free(&node->v.i);
		break;
	}
}

void bfloop_tree_free(bfloop_node *root)
{
	if (root == NULL)
		return;
	bfloop_tree_free(root->children);
	bfloop_tree_free(root->siblings);
	bfloop_node_free(root);
	free(root);
}

static void bfloop_node_print(const bfloop_node *node)
{
	wchar_t *s;

	switch (node->type) {
	case NODE_NONE:
		printf("NONE");
		break;
	case NODE_NO:
		printf("NO");
		break;
	case NODE_YES:
		printf("YES");
		break;
	case NODE_INT:
		s = bigint_render(&node->v.i);
		printf("INT<%ls>", s);
		free(s);
		break;
	case NODE_IDENT:
		printf("IDENT<%ls>", node->v.s);
		break;
	case NODE_CELL_REF:
		s = bigint_render(&node->v.i);
		printf("CELL<%ls>", s);
		free(s);
		break;
	case NODE_OUTPUT:
		printf("OUTPUT");
		break;
	case NODE_ADD:
		printf("ADD");
		break;
	case NODE_MUL:
		printf("MUL");
		break;
	case NODE_EQUALS:
		printf("EQUALS");
		break;
	case NODE_LESS_THAN:
		printf("LESS_THAN");
		break;
	case NODE_GREATER_THAN:
		printf("GREATER_THAN");
		break;
	case NODE_AND:
		printf("AND");
		break;
	case NODE_ASSIGN:
		printf("ASSIGN");
		break;
	case NODE_CALL:
		printf("CALL<%ls>", node->v.s);
		break;
	case NODE_LOOP:
		printf("LOOP");
		break;
	case NODE_MU_LOOP:
		printf("MU-LOOP");
		break;
	case NODE_ABORT:
		s = bigint_render(&node->v.i);
		printf("ABORT<%ls>", s);
		free(s);
		break;
	case NODE_QUIT:
		s = bigint_render(&node->v.i);
		printf("QUIT<%ls>", s);
		free(s);
		break;
	case NODE_IF:
		printf("IF");
		break;
	case NODE_BLOCK:
		s = bigint_render(&node->v.i);
		printf("BLOCK<%ls>", s);
		free(s);
		break;
	case NODE_PROCEDURE:
		printf("PROCEDURE<%ls>", node->v.s);
		break;
	case NODE_TOPLEVEL:
		printf("TOPLEVEL");
		break;
	}
}

static void bfloop_tree_print_depth(const bfloop_node *root, int depth)
{
	int i;

	if (root == NULL)
		return;

	for (i = 0; i < depth; i++)
		printf(". ");
	bfloop_node_print(root);
	printf("\n");
	bfloop_tree_print_depth(root->children, depth + 1);
	bfloop_tree_print_depth(root->siblings, depth);
}

void bfloop_tree_print(const bfloop_node *root)
{
	bfloop_tree_print_depth(root, 0);
}

static void indent(int depth)
{
	while (depth-- > 0)
		printf("\t");
}

static void bfloop_tree_reconstitute_depth(const bfloop_node *root, int depth)
{
	const bfloop_node *p;
	wchar_t *s;

	if (root == NULL)
		return;

	switch (root->type) {
	case NODE_TOPLEVEL:
		internal_assert(root->siblings == NULL);
		for (p = root->children; p != NULL; p = p->siblings) {
			bfloop_tree_reconstitute_depth(p, depth);
			printf("\n");
			if (p->siblings != NULL && p->type == NODE_PROCEDURE)
				printf("\n");
		}
		break;
	case NODE_PROCEDURE:
		indent(depth);
		printf("DEFINE PROCEDURE ''%ls'' [", root->v.s);
		internal_assert(root->children != NULL);
		for (p = root->children->siblings; p != NULL; p = p->siblings) {
			bfloop_tree_reconstitute_depth(p, 0);
			if (p->siblings != NULL)
				printf(", ");
		}
		printf("]:\n");
		bfloop_tree_reconstitute_depth(root->children, depth);
		printf(".");
		break;
	case NODE_BLOCK:
		indent(depth);
		s = bigint_render(&root->v.i);
		internal_assert(s != NULL);
		printf("BLOCK %ls: BEGIN\n", s);
		for (p = root->children; p != NULL; p = p->siblings) {
			bfloop_tree_reconstitute_depth(p, depth + 1);
			printf(";\n");
		}
		indent(depth);
		printf("BLOCK %ls: END", s);
		free(s);
		break;
	case NODE_LOOP:
		indent(depth);
		printf("LOOP ");
		bfloop_tree_reconstitute_depth(root->children, 0);
		printf(" TIMES:\n");
		bfloop_tree_reconstitute_depth(root->children->siblings, depth);
		break;
	case NODE_MU_LOOP:
		indent(depth);
		printf("MU-LOOP:");
		bfloop_tree_reconstitute_depth(root->children, depth);
		break;
	case NODE_IF:
		indent(depth);
		printf("IF ");
		bfloop_tree_reconstitute_depth(root->children, 0);
		printf(", THEN:\n");
		bfloop_tree_reconstitute_depth(root->children->siblings, depth);
		break;
	case NODE_QUIT:
		indent(depth);
		s = bigint_render(&root->v.i);
		internal_assert(s != NULL);
		printf("QUIT BLOCK %ls", s);
		free(s);
		break;
	case NODE_ABORT:
		indent(depth);
		s = bigint_render(&root->v.i);
		internal_assert(s != NULL);
		printf("ABORT LOOP %ls", s);
		free(s);
		break;
	case NODE_CALL:
		indent(depth);
		printf("%ls [", root->v.s);
		for (p = root->children; p != NULL; p = p->siblings) {
			bfloop_tree_reconstitute_depth(p, 0);
			if (p->siblings != NULL)
				printf(",");
		}
		printf("]");
		break;
	case NODE_ASSIGN:
		indent(depth);
		bfloop_tree_reconstitute_depth(root->children, 0);
		printf(" %ls ", L"⇐");
		internal_assert(root->children != NULL);
		bfloop_tree_reconstitute_depth(root->children->siblings, 0);
		break;
	case NODE_AND:
		internal_assert(root->children != NULL);
		internal_assert(root->children->siblings != NULL);
		indent(depth);
		printf("{");
		bfloop_tree_reconstitute_depth(root->children, 0);
		printf(" AND ");
		bfloop_tree_reconstitute_depth(root->children->siblings, 0);
		printf("}");
		break;
	case NODE_EQUALS:
		indent(depth);
		bfloop_tree_reconstitute_depth(root->children, 0);
		printf(" %ls ", L"=");
		internal_assert(root->children != NULL);
		bfloop_tree_reconstitute_depth(root->children->siblings, 0);
		break;
	case NODE_GREATER_THAN:
		indent(depth);
		bfloop_tree_reconstitute_depth(root->children, 0);
		printf(" %ls ", L">");
		internal_assert(root->children != NULL);
		bfloop_tree_reconstitute_depth(root->children->siblings, 0);
		break;
	case NODE_LESS_THAN:
		indent(depth);
		bfloop_tree_reconstitute_depth(root->children, 0);
		printf(" %ls ", L"<");
		internal_assert(root->children != NULL);
		bfloop_tree_reconstitute_depth(root->children->siblings, 0);
		break;
	case NODE_ADD:
		indent(depth);
		bfloop_tree_reconstitute_depth(root->children, 0);
		printf(" %ls ", L"+");
		internal_assert(root->children != NULL);
		bfloop_tree_reconstitute_depth(root->children->siblings, 0);
		break;
	case NODE_MUL:
		indent(depth);
		bfloop_tree_reconstitute_depth(root->children, 0);
		printf(" %ls ", L"×");
		internal_assert(root->children != NULL);
		bfloop_tree_reconstitute_depth(root->children->siblings, 0);
		break;
	case NODE_CELL_REF:
		indent(depth);
		s = bigint_render(&root->v.i);
		internal_assert(s != NULL);
		printf("CELL(%ls)", s);
		free(s);
		break;
	case NODE_OUTPUT:
		indent(depth);
		printf("OUTPUT");
		break;
	case NODE_IDENT:
		indent(depth);
		printf("%ls", root->v.s);
		break;
	case NODE_INT:
		indent(depth);
		s = bigint_render(&root->v.i);
		internal_assert(s != NULL);
		printf("%ls", s);
		free(s);
		break;
	case NODE_YES:
		indent(depth);
		printf("YES");
		break;
	case NODE_NO:
		indent(depth);
		printf("NO");
		break;
	case NODE_NONE:
		internal_error("NONE node in tree.");
		break;
	}
}

/* Print out the reconstituted text of a program given a parse tree. */
void bfloop_tree_reconstitute(const bfloop_node *root)
{
	bfloop_tree_reconstitute_depth(root, 0);
}
